/*************************************************************************************************
*
*
*			ObjectMacZapp		-- a standard Mac OOP application template
*
*
*
*			ZPaintEngine.h		-- an object for implementing paint functions in a window
*
*
*
*
*
*			 1996, Graham Cox
*
*
*
*
*************************************************************************************************/


#pragma once

#ifndef	__ZPAINTENGINE__
#define __ZPAINTENGINE__


#include	"ZComrade.h"
#include	"ZGWorld.h"



/*

This object implements the core paint functions of a paint-type program. It requires the use
of the ZGWorld class for buffering the operations. Additional work is required to implement
an interactive user interface, though not much. The workhorse of the class is MouseClick()- call
it with the mouse point mapped to the artwork coordinates, and everything will mostly just
work- you'll get painting, dragging, fills, etc according to the state variables of the engine.
The artwork is blitted to the port that is current when MouseClick is first called- usually this
will be a window. If the window is displaying the artwork in a zoomed or scrolled fashion, the
caller must translate the mouse coordinates as necessary. To facilitate accurate blitting when
drawing interactively, the artwork is mapped to the window according to the value set up in
the artBounds rectangle. It is up to the caller to set this up properly.

*/

// forward class references

class	ZUndoPaintTask;
class	ZGradient;

// types of tools

typedef enum
{
	noTool = 0,					// 0
	paintBrushTool,				// 1
	airbrushTool,				// 2
	pencilTool,					// 3
	lineTool,					// 4
	eraserTool,					// 5
	fillTool,					// 6
	textTool,					// 7
	rectFrameTool,				// 8
	rectSolidTool,				// 9
	roundRectFrameTool,			// 10
	roundRectSolidTool,			// 11
	ovalFrameTool,				// 12
	ovalSolidTool,				// 13
	arcFrameTool,				// 14
	arcSolidTool,				// 15
	regularPolyFrameTool,		// 16
	regularPolySolidTool,		// 17
	irregPolyFrameTool,			// 18
	irregPolySolidTool,			// 19
	stampTool,					// 20
	colourPickTool,				// 21
	sprayCanTool,				// 22
	invertTool,					// 23
	zoomTool,					// 24
	selectionTool				// 25
}
PaintTool;

// types of selection tool:

typedef enum
{
	rectSelection = 1,
	ovalSelection,
	freeSelection,
	lassoSelection,
	wandSelection
}
SelectionType;

// selection modifiers:

typedef enum
{
	replaceSelection = 0,
	addToSelection,
	subtractFromSelection,
	shrinkSelection = 8,
	moveSelection = 16,
	moveSelectionWithImage ,
	moveSelectionWithImageCopy
}
SelectionModifier;

// types of fill applied by the fill tool

typedef enum
{
	noFill = 0,
	plainColourFill,
	patternFill,
	colourPatternFill,
	gradientFill
}
FillType;

// paint types

typedef enum
{
	colourPaint = 1,
	patternPaint,
	colourPatternPaint
}
PaintType;

// brush types

enum
{
	ovalBrush = 1,
	rectBrush,
	slantLeftBrush,
	slantRightBrush,
	bitMapBrush,
	spraycanBrush,
	airBrush
};

typedef short BrushType;

// airbrush modes

enum
{
	stdAirMode = 1,
	lightenArt,
	darkenArt,
	stdSplatter,
	multiSplatter
};

typedef short AirMode;

// selector constants

enum
{
	kErase,
	kFrame,
	kFill,
	kTextFill
};

// messages

enum
{
	peMsgNewPaintColourPicked	= 'pepc',
	peMsgNewEraseColourPicked	= 'peec',
	peMsgNewLineColourPicked	= 'pelc',
	peMsgSelectionRgnChanged	= 'pesx',
	peMsgTextEditingBegun		= 'pet+',
	peMsgTextEditingEnded		= 'pet-',
	peMsgTextParamsChanged		= 'petx',
	peMsgFillStyleChanged		= 'pefx',
	peMsgMouseLocationInfo		= 'pemi'
};

// struct for saving & restoring the state of the paint engine (in prefs, etc.) Note that
// things such as the selection region are NOT saved here- this can only restore paint
// parameters in general.

#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=mac68k
#endif

typedef struct
{
	short			sVersNum;
	Boolean			aaLines : 1;
	Boolean			aaText : 1;	
	Boolean			tween : 1;
	Boolean			frTsp : 1;
	Boolean			fillTsp : 1;
	Boolean			paintTsp : 1;
	Boolean			spare : 2;
	short			slipTolerance;
	short			frWidth;
	short			frHeight;
	short			cPatFillID;
	short			cPatFrameID;
	short			brWidth;		
	short			brHeight;
	short			airRadius;
	short			airPressure;
	short			txShadowWidth;
	short			txShadowHeight;
	short			txFont;
	short			txSize;
	short			txStyle;
	short			txBorder;
	short			rPolySides;
	short			rRectHCorner;
	short			rRectVCorner;
	short			curGradID;
	RGBColor		paint;
	RGBColor		lines;
	RGBColor		erase;
	SelectionType	curSelType;
	PaintTool		curTool;
	FillType		curFill;
	PaintType		curPaint;		
	BrushType		curBrush;		
	AirMode			curAirMode;
	FillType		txFill;
	Pattern			pFill;
	Pattern			pFrame;
	long			reserved[4];
}
PEStateRec, *PEStatePtr, **PEStateHdl;


#if PRAGMA_ALIGN_SUPPORTED
#pragma options align=reset
#endif

// class definition:

class	ZPaintEngine	: public ZComrade
{
protected:
	ZGWorld*			artBuffer;			// buffer containing artwork
	ZGWorld*			tempBuffer;			// utility buffer used internally
	ZGWorld*			airGW;				// offscreen used for some airbrush effects
	ZGWorld*			floatBuffer;		// offscreen for moving image selections.
	Boolean				ownsArtBuffer;		// TRUE if we dispose art buffer when deleted
	PaintTool			currentTool;		// current tool selected
	Boolean				antiAlias;			// TRUE for anti-aliased lines and shapes
	Boolean				antiAliasText;		// TRUE for anti-aliased text
	Boolean				drawFromCentre;		// TRUE if elastic shapes expand from centre
	Boolean				tweenEnable;		// TRUE if paintbrush is tweened
	Boolean				constrain;			// TRUE if shape is constrained
	GrafPtr				windowPort;			// display port drawing appears in
	Rect				artFrame;			// true dimensions of artwork
	Rect				artBounds;			// displayed dimensions of artwork (to allow zoom, etc)
	short				hAlign;				// point alignment factor to allow for scaling & offset
	short				vAlign;				// same for vertical dimension	
	Point				cloneAnchor;		// anchor point for clone tool
	short				cloneHOffset;		// h offset from anchor to draw point
	short				cloneVOffset;		// v offset from anchor to draw point
	Boolean				hadFirstCloneClick;	// had first subsequent click after setting anchor

// selection members:

	RgnHandle			theSelection;		// selected area in image ("marching ants")
	SelectionType		currentSelType;		// current method for selecting
	SelectionModifier	currentSelMod;		// current selection modifier
	char				ants[8];			// marching ants pattern
	short				fbHOffset;			// horizontal position of float buffer
	short				fbVOffset;			// vertical position of float buffer
	short				matchTolerance;		// colour slop factor for fill, wand and lasso
	Boolean				antsActive;			// TRUE if ants animating
	
// mask region members:

	RgnHandle			paintMask;			// mask region
	Boolean				paintMaskActive;	// TRUE if mask is applied
	Boolean				paintMaskVisible;	// TRUE if mask is visible
	
// text handling members:
	
	TEHandle			currTE;				// editing record
	Boolean				inTextMode;			// TRUE if text edition in progress
	Rect				textRect;			// text editing rect
	
// paint parameters:

	short				frameWidth;			// width of lines and frames
	short				frameHeight;		// height of lines and frames
	RGBColor			paintColour;		// fill colour
	RGBColor			lineColour;			// line and frame colour
	RGBColor			eraseColour;		// background and erase colour
	Pattern				fillPattern;		// pattern for fills
	Pattern				framePattern;		// pattern for lines and frames
	PixPatHandle		cFillPattern;		// colour fill pattern
	PixPatHandle		cFramePattern;		// colour line pattern
	FillType			currentFill;		// type of fill for fill tool
	PaintType			currentPaint;		// type of paint for paintbrush
	BrushType			currentBrush;		// type of brush we're painting with	
	short				brushWidth;			// width of paintbrush
	short				brushHeight;		// height of paintbrush
	BitMap				brushBits;			// brush bitmap currently set
	Boolean				frameBkTransparent;	// back colour of frame pat is transparent
	Boolean				fillBkTransparent;	// back colour of fill pattern is transparent
	Boolean				paintBkTransparent;	// back colour of paint is transparent
	Boolean				txFillTransparent;	// text fill is transparent
	AirMode				currentAirMode;		// airbrush style
	short				airRadius;			// radius of airbrush spray
	short				airPressure;		// "force" of spray
	short				textShadowWidth;	// h offset of text shadow
	short				textShadowHeight;	// v offset of text shadow
	short				textFont;			// font ID to use
	short				textSize;			// text size
	short				textStyle;			// style of text
	short				textAlign;			// alignment of text
	FillType			textFill;			// fill style for text
	short				textBorderWidth;	// frame border for text, if any
	short				regularPolySides;	// number of sides for regular polygons
	short				roundRectHCorner;	// horiz corner dimension for roundrect
	short				roundRectVCorner;	// vert corner dimension for roundrect
	ZGradient*			curGradient;		// current gradient fill selected
	ZUndoPaintTask*		curPTask;			// current undo task built
	
public:

	ZPaintEngine( ZGWorld* anArtBuffer );
	ZPaintEngine( const Rect& artworkSize );
	ZPaintEngine();
	~ZPaintEngine();

// basic setting up
	
	virtual void		SetDefaults();
	virtual void		SetCurrentTool( const PaintTool aTool );
	virtual void		SetSelectionType( const SelectionType aType );
	virtual void		SetArtBounds( const Rect& aRect, const short scaling = 100 );
	virtual void		SetArtBuffer( ZGWorld* aBuf, Boolean ownsBuf = FALSE );
	
// state handles:

	virtual void		SetStateFromHandle( PEStateHdl pesH );
	virtual PEStateHdl	GetStateHandle();

// setting paint parameters:

	virtual void		SetPaintBrush( const BrushType aType );
	virtual void		SetPaintType( const PaintType ptType );
	virtual void		SetPaintBrushSize( const short aWidth, const short aHeight );
	virtual void		SetPaintColour( const RGBColor& aColour );
	virtual void		SetFrameColour( const RGBColor& aColour );
	virtual void		SetEraseColour( const RGBColor& aColour );
	virtual void		SetPaintPattern( const Pattern& aPat );
	virtual void		SetPaintPattern( const short patListID, const short patID );
	virtual void		SetFramePattern( const Pattern& aPat );
	virtual void		SetFramePattern( const short patListID, const short patID );
	virtual void		SetPaintCPattern( PixPatHandle pp );
	virtual void		SetFrameCPattern( PixPatHandle pp );
	
// setting fill parameters:

	virtual void		SetFillType( const FillType aType );
	virtual void		SetGradientFill( ZGradient* aGradient );
	
// airbrush params:

	virtual void		SetAirbrushType( const AirMode abType );
	virtual void		SetAir( const short aRadius, const short aPressure );
	
// text params:

	virtual void		SetTextParams( const short aFont, const short aSize, const short aStyle, const short align );
	virtual void		SetTextShadow( const short shadowH, const short shadowV );
	virtual void		SetTextBorderFrame( const short bfWidth );
	
// polygons

	virtual void		SetRegularPolySides( const short sides );
	
// the demon high level call for processing mouse clicks in the image:

	virtual void		MouseClick( Point mClickPt, const short modifiers );
	
// other setters:

	inline	void		SetFrameWidth( short aWidth ) { frameWidth = frameHeight = aWidth; };
	inline	void		SetRoundCornerSize( Point aCorner ) { roundRectHCorner = aCorner.h; roundRectVCorner = aCorner.v; };
	inline	void		SetBrushBitMap( BitMap* aBits ) { brushBits = *aBits; };
	inline	void		SetTweening( Boolean aTween ) { tweenEnable = aTween; };
	inline	void		SetFillTransparent( Boolean isTsp ) { fillBkTransparent = isTsp; };
	inline	void		SetPaintTransparent( Boolean isTsp ) { paintBkTransparent = isTsp; };
	inline	void		SetPortTextParams() { TextFont( textFont ); TextSize( textSize ); TextFace( textStyle ); };
	inline	void		SetTextFill( FillType aType ) { textFill = aType; };
	inline  void		SetTextTransparent( Boolean isTsp ) { txFillTransparent = isTsp; };
	
	inline 	void		SetCMatchTolerance( const short tol ) { matchTolerance = tol; };
	inline  short		GetCMatchTolerance() { return matchTolerance; };
	
// getters:

	inline	PaintTool		GetPaintTool()	{ return currentTool; };
	inline	BrushType		GetBrushType()	{ return currentBrush; };
	inline	PaintType		GetPaintType()	{ return currentPaint; };
	inline	FillType		GetFillType()	{ return currentFill; };
	inline	SelectionType	GetSelectionType() { return currentSelType; };
	inline	void			SetWindowPort( GrafPtr aPort ) { windowPort = aPort; };
	inline	ZGradient*		GetGradient()	{ return curGradient; };
	inline	void			GetSelection( RgnHandle aRgn ) { CopyRgn( theSelection, aRgn ); };
	inline	void			GetArtFrame( Rect* aRect ) { *aRect = artFrame; };
	
	inline	Pattern			GetBWFillPattern() { return fillPattern; };
	inline	PixPatHandle	GetCFillPattern() { return cFillPattern; };
	inline	void			GetPaintColour( RGBColor* aColour ) { *aColour = paintColour; };
	inline	void			GetLineColour( RGBColor* aColour ) { *aColour = lineColour; };
	inline  void			GetEraseColour( RGBColor* aColour ) { *aColour = eraseColour; };
	inline	Boolean			GetPaintTransparent() { return paintBkTransparent; };
	inline	short			GetFrameWidth() { return frameWidth; };
	inline	short			GetFrameHeight() { return frameHeight; };
	inline	short			GetBrushWidth() { return brushWidth; };
	inline	short			GetBrushHeight() { return brushHeight; };
	inline	short			GetAirbrushRadius() { return airRadius; };
	inline  short			GetAirbrushPressure() { return airPressure; };
	inline  AirMode			GetAirbrushMode() { return currentAirMode; };
	inline  short			GetRoundRectWidth() { return roundRectHCorner; };
	inline	short			GetRoundRectHeight() { return roundRectVCorner; };
	inline	short			GetRegularPolySides() { return regularPolySides; };
	
	inline	Boolean			InTextMode() { return inTextMode; };
	inline	short			GetTextBorderFrame() { return textBorderWidth; };
	inline	void			GetTextAttribs( FillType* ft, Boolean* isTsp ) { *ft = textFill; *isTsp = txFillTransparent; };
	
	virtual	CTabHandle		GetPaintColours();	
	
// public methods for selection & text access:

	virtual void		AnimateSelection();
	inline Boolean		HasSelection() { return !EmptyRgn( theSelection ); };
	virtual Boolean		PtInSelection( const Point mouse );
	
	virtual void		UpdateFloatToWindow();
	virtual void		ClearSelection();
	virtual void		CopySelection();
	virtual void		PasteSelection();
	virtual void		ActivateSelection( Boolean state );
	virtual void		SetSelection( RgnHandle aRgn, Boolean killFB = TRUE );
	virtual void		SelectAll();
	virtual void		Clear();
	virtual void		MakeFloatingSelection( Boolean eraseBehind = FALSE );
	virtual void		FillSelection( Point centre, RgnHandle aRgn = NULL );
	
// text rendering (called by DRAWHook- do not use directly)

	virtual void		RenderText( Ptr textBuf, unsigned short charOffset, unsigned short chCount );
	virtual void		TextType( char theChar );
	virtual void		EndText();

// we can use the paint engine to fill regions externally, so this method is public	
	
	virtual void		FillRegionInPort( RgnHandle aRgn, Point centre );
	virtual void		SetPortPenPattern( const short which );

// regular polygon calculation:

	virtual RgnHandle	MakeRegularPoly( Point centre, Point aVertex );

// effects:

	virtual void		Flip( const short flipDir );
	virtual void		Invert();

// drawing primitive operations:

	virtual void		DrawPaint( Point location );
	virtual void		DrawShapeInPort( Rect* aRect );
	virtual void		DrawShapeInPort( RgnHandle aRgn );
	virtual void		DrawLineInPort( Point a, Point b );
	virtual void		DrawAirInPort( Point location, Rect* artSource );
	virtual void		DrawArcInPort( Point anchor, Point location );
	virtual void		PickUpPaintColour( Point location, short modifiers );
	virtual void		PickUpPaintColour( RGBColor* aColour , short modifiers );
	
// paint mask manipulation:
	
	virtual void		SetMaskToSelection();
	virtual void		AddToMask();
	virtual void		SubtractFromMask();
	virtual void		ClearMask();
	virtual void		InvertMask();
	virtual void		ShowHideMask( Boolean showIt );
	virtual void		ActivateMask( Boolean isActive );
	virtual void		RefreshMask();
	inline	Boolean		MaskActive() { return paintMaskActive; };
	inline	Boolean		MaskVisible() { return paintMaskVisible; };
	
	virtual void		PrepareMaskClipForPort();

protected:

// double-buffering utils:

	virtual void		SetUpSecondBuffer( Rect* aRect, Boolean autoClear = TRUE );
	virtual void		RemoveSecondBuffer();
	virtual void		MakeBitMap( Rect* aRect, BitMap* aBMap );
	virtual void		DisposeBitMap( BitMap* aBMap );
	
// lower level mouse tracking, etc.

	virtual void		DragElasticShape( Point mouseStart );
	virtual void		DragLineOrPoly( Point mouseStart, Boolean isPoly = FALSE );
	virtual void		PaintTrail( Point mouseStart );
	virtual void		AirbrushTrail( Point mouseStart );
	virtual void		ClickPaint( Point mouseStart );
	virtual void		GetClickedColour( Point mouse, RGBColor* aColour );
	virtual void		TweenPaint( Point start, Point end );
	
// calculating the seed fill region:

	virtual void		SeedFillToRegion( Point clickPt, RgnHandle aRgn );
	
// port utils:

	virtual void		BlitToWindow( Rect* srcRect, Rect* destRect, ZGWorld* aBuffer = NULL );
	virtual void		UpdateWindow();

// undo task:
	
	virtual ZUndoPaintTask*	MakeUndoTask( short taskID = 0 );
	virtual void		SendUndoTask();	

// point translation:

public:	
	virtual void		TranslatePointToOffscreen( Point* aPt );
	virtual void		TranslateRectToOffscreen( Rect* aRect );
	virtual void		ReportMouseLoc( Point anchor, Point loc, Boolean inDrag );
	
// low-level text handling:

protected:
	virtual void		BeginText( Point mouseStart );
	virtual void		TextClick( Point mousePt, short modifiers );
	virtual void		DragTextRect( Point startPt );
	virtual void		CopyTextBufToWindow();
	
// handling selections:

	virtual void		DragSelection( Point mouseStart, short modifiers );
	virtual void		KillFloatingSelection( Boolean copyBack = TRUE );
	virtual void		ShrinkSelection( Point cSeedPt, RgnHandle aRgn );
	virtual void		InvalSelection();
};

#define		kPaintingTaskStrListID		201
#define		kNoUndoAvailableAlertID		201
#define		kEraserToolHalfWidth		5

// utility function:

Boolean		EqualColour( RGBColor* a, RGBColor* b );

// additional task IDs not selected by tool

enum
{
	kTaskMove = 21,
	kTaskPlaceGraphic,
	kTaskClear = 24
};


#endif